#include <WiFi.h>
#include <PubSubClient.h>
#include <Adafruit_NeoPixel.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

// WiFi & MQTT Config
const char* ssid = "wifiname";
const char* password = "wifipassword!";
const char* mqtt_server = "rpi ip";

WiFiClient espClient;
PubSubClient client(espClient);

// LED Strip Config
#define LED_PIN 13
#define LED_COUNT 16
Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);

// OLED Config
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define OLED_RESET    -1
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// Lamp State
const int LAMP_ID = 1;
bool isSelected = false;
bool isOn = false;
int currentMode = 0;
const int TOTAL_MODES = 4;

// Timing
unsigned long lastGestureTime = 0;
const unsigned long GESTURE_COOLDOWN = 3000;  // ms

String lastModeGesture = "";

// Color Struct
struct RGBColor {
  uint8_t r, g, b;
};

RGBColor getColorForMode(int mode) {
  switch (mode) {
    case 0: return {255, 255, 255}; // White
    case 1: return {0, 120, 255};   // Like (Blue)
    case 2: return {255, 50, 50};   // Dislike (Red)
    case 3: return {0, 255, 100};   // Vsign (Green)
    default: return {0, 0, 0};
  }
}

void fadeToColor(uint8_t r, uint8_t g, uint8_t b, int steps = 30, int delayTime = 20) {
  for (int s = 0; s <= steps; s++) {
    for (int i = 0; i < strip.numPixels(); i++) {
      uint32_t currentColor = strip.getPixelColor(i);
      uint8_t cr = (currentColor >> 16) & 0xFF;
      uint8_t cg = (currentColor >> 8) & 0xFF;
      uint8_t cb = currentColor & 0xFF;

      strip.setPixelColor(i, strip.Color(
        cr + (r - cr) * s / steps,
        cg + (g - cg) * s / steps,
        cb + (b - cb) * s / steps
      ));
    }
    strip.show();
    delay(delayTime);
  }
}

void updateDisplay(const String& gesture = "") {
  display.clearDisplay();

  // Small text: lamp ID
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.print("Lamp 1");

  // Big text: mode name + number
  display.setTextSize(2);
  display.setCursor(0, 20);

  // Show mode name and number
  String modeName;
  switch (currentMode) {
    case 0: modeName = "White"; break;
    case 1: modeName = "Like"; break;
    case 2: modeName = "Dislike"; break;
    case 3: modeName = "Vsign"; break;
    default: modeName = "Unknown";
  }

  display.print("Mode ");
  display.print(currentMode);
  display.print(": ");
  display.println(modeName);

  display.display();
}

void setMode(int mode) {
  currentMode = mode % TOTAL_MODES;
  RGBColor color = getColorForMode(currentMode);
  fadeToColor(color.r, color.g, color.b);
  updateDisplay();
}

void turnOffLamp() {
  fadeToColor(0, 0, 0);
  isOn = false;
  updateDisplay();
}

void turnOnLamp() {
  isOn = true;
  setMode(currentMode);
}

void mqttCallback(char* topic, byte* payload, unsigned int length) {
  String message;
  for (unsigned int i = 0; i < length; i++) {
    message += (char)payload[i];
  }

  String selectTopic = "rpi/lamp/select";
  String lampTopic = "lamp1/broadcast";

  if (String(topic) == selectTopic) {
    isSelected = (message.toInt() == LAMP_ID);
    Serial.print(message);
    updateDisplay();
    return;
  }

  if (String(topic) == lampTopic) {
    unsigned long now = millis();
    if (now - lastGestureTime < GESTURE_COOLDOWN) return;
    lastGestureTime = now;
    Serial.print(message);

    if (message == "palm") {
      isOn = !isOn;
      if (isOn){
        turnOnLamp();
      }
      else{ 
        turnOffLamp();
      }
    }
    else if (message == "like" || message == "dislike" || message == "vsign") {
      if (message != lastModeGesture) {
        if (message == "like") currentMode = 1;
        else if (message == "dislike") currentMode = 2;
        else if (message == "vsign") currentMode = 3;

        if (isOn) setMode(currentMode);
        updateDisplay(message);

        lastModeGesture = message;
      }
    }
  }
}

void setupWiFi() {
  WiFi.begin(ssid, password);
  Serial.print("Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\nWiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void reconnectMQTT() {
  while (!client.connected()) {
    Serial.print("Connecting to MQTT...");
    if (client.connect(("LampClient" + String(LAMP_ID)).c_str())) {
      Serial.println("connected");
      client.subscribe("lamp1/broadcast");
      client.subscribe("rpi/lamp/select");
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 2 seconds");
      delay(2000);
    }
  }
}

void setup() {
  Serial.begin(115200);

  strip.begin();
  strip.setBrightness(255);
  strip.show();

  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    Serial.println("SSD1306 allocation failed");
    while (true);
  }

  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("Waiting for");
  display.println("gesture...");
  display.display();

  setupWiFi();
  client.setServer(mqtt_server, 1883);
  client.setCallback(mqttCallback);
}

void loop() {
  if (!client.connected()) reconnectMQTT();
  client.loop();
}
